from base_handler import BaseHandler
from datetime import datetime
from typing import List, Dict, Any, Optional
import json
import re
import os
import sys

# Define crawl_website as None for import failure check
crawl_website = None

try:
    import requests
except ImportError:
    print("Please install required packages: pip install requests")

# Try multiple import methods
try:
    from ..web.crawl import crawl_website  # Import crawl_website function
except ImportError as e:
    try:
        # Try adding parent directory to path
        parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        if parent_dir not in sys.path:
            sys.path.insert(0, parent_dir)
        from pm_studio_mcp.utils.web.crawl import crawl_website
    except ImportError:
        print(f"Warning: Could not import crawl_website from crawl_utils: {e}")
        # Fallback function if import fails
        def crawl_website(url: str) -> str:
            """Fallback crawl_website function using requests."""
            try:
                import requests
                response = requests.get(url, timeout=30)
                response.raise_for_status()
                return response.text
            except Exception as e:
                raise Exception(f"Error fetching URL {url}: {str(e)}")

class ChromeFeaturesHandler(BaseHandler):
    """
    Handler for Chrome features operations.
    This class is responsible for fetching Chrome version features from Chrome Status API.
    """
    _instance = None
    _initialized = False

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(ChromeFeaturesHandler, cls).__new__(cls)
        return cls._instance
    
    def __init__(self):
        """Initialize the ChromeFeaturesHandler."""
        if self._initialized:
            return
        
        self.base_url = "https://chromestatus.com/api/v0/features"
        self.chrome_version_url = "https://chromestatus.com/api/v0/channels"
        self._initialized = True
        
        # Check if crawl_website is available
        if crawl_website is None:
            print("Warning: crawl_website function is not available")

    def get_current_chrome_version(self) -> Optional[int]:
        """
        Get the current stable Chrome version number.
        
        Returns:
            int: Current Chrome stable version milestone number
        """
        try:
            # Check if crawl_website is available
            if crawl_website is None:
                raise Exception("crawl_website function is not available")
            
            # Use crawl_website to get version info
            response_text = crawl_website(self.chrome_version_url)
            
            # Clean up possible special characters
            response_text = self._clean_json_response(response_text)
            
            # Parse JSON
            channels_data = json.loads(response_text)
            
            # Handle new data structure: dictionary form, keys are channel names
            if isinstance(channels_data, dict):
                # Check if 'stable' key exists
                if 'stable' in channels_data:
                    stable_data = channels_data['stable']
                    if isinstance(stable_data, dict):
                        # Get milestone from stable data
                        milestone = stable_data.get('milestone')
                        if milestone is not None:
                            try:
                                return int(milestone)
                            except (ValueError, TypeError):
                                pass
                        
                        # If no milestone, try to extract from version
                        version = stable_data.get('version')
                        if version is not None:
                            # version could be int or str
                            if isinstance(version, (int, float)):
                                return int(version)
                            elif isinstance(version, str):
                                # Extract the first part of the version as milestone
                                # For example "131.0.6778.85" -> 131
                                version_parts = version.split('.')
                                if version_parts:
                                    try:
                                        return int(version_parts[0])
                                    except ValueError:
                                        pass
                        
                        # Print stable_data content for debugging
                        print(f"Stable channel data: {stable_data}")
                        print(f"Available fields in stable: {list(stable_data.keys())}")
                
                # If all above methods fail, try to iterate over all channels
                for channel_name, channel_data in channels_data.items():
                    if channel_name.lower() == 'stable' and isinstance(channel_data, dict):
                        milestone = channel_data.get('milestone')
                        if milestone is not None:
                            try:
                                return int(milestone)
                            except (ValueError, TypeError):
                                pass
            
            # Handle case where data might be a list (old format)
            elif isinstance(channels_data, list):
                # Look for stable version
                for channel in channels_data:
                    if isinstance(channel, dict) and channel.get('channel') == 'stable':
                        milestone = channel.get('milestone')
                        if milestone is not None:
                            try:
                                return int(milestone)
                            except (ValueError, TypeError):
                                pass
            
            # If not found, print data structure for debugging
            print(f"Could not find stable version in data structure: {type(channels_data)}")
            print(f"Available keys: {list(channels_data.keys()) if isinstance(channels_data, dict) else 'N/A'}")
            
            return None
            
        except Exception as e:
            print(f"Error getting current Chrome version: {str(e)}")
            import traceback
            traceback.print_exc()
            return None

    def _clean_json_response(self, response_text: str) -> str:
        """
        Clean the JSON response by removing problematic characters.
        
        Args:
            response_text: Raw response text
            
        Returns:
            str: Cleaned JSON text
        """
        # Remove possible ")]}'" prefix
        if response_text.startswith(")]}'\n"):
            response_text = response_text[5:]
        elif response_text.startswith(")]}'"):
            response_text = response_text[4:]
        
        # Extract content within body tag (if exists)
        body_match = re.search(r'<body[^>]*>(.*?)</body>', response_text, re.DOTALL | re.IGNORECASE)
        if body_match:
            response_text = body_match.group(1)
        
        # Remove other HTML tags
        response_text = re.sub(r'<[^>]+>', '', response_text)
        
        return response_text.strip()

    def fetch_data(self, product_name, start_date=None, end_date=None, *args, **kwargs):
        """
        Fetch Chrome features data based on product name, date range, and version.
        
        Args:
            product_name: Name of the product (should be 'Chrome')
            start_date: Optional start date to look for Chrome versions
            end_date: Optional end date to look for Chrome versions
            *args: Variable length argument list
            **kwargs: Arbitrary keyword arguments including:
                - milestone: Specific Chrome version milestone number
                - versions: List of Chrome versions
                - working_dir: Directory to save output files
                - category_filter: List of categories to filter
                - status_filter: List of statuses to filter
            
        Returns:
            Dict containing status and results
        """
        if not product_name:
            raise ValueError("Product name is required.")
        
        # Check if the product is Chrome
        if product_name.lower() != "chrome":
            return {
                "status": "error",
                "message": f"This handler only supports Chrome. Received: {product_name}"
            }
        
        # Extract parameters from kwargs
        working_dir = kwargs.get("working_dir", "")
        category_filter = kwargs.get("category_filter")
        status_filter = kwargs.get("status_filter")
        
        # Get milestone(s) from kwargs
        milestone = kwargs.get("milestone")
        versions = kwargs.get("versions", [])
        
        # If versions are provided but no milestone is set, use the first version as milestone
        if not milestone and versions:
            if isinstance(versions, list) and versions:
                milestone = versions[0]
            elif isinstance(versions, (int, str)):
                milestone = versions
        
        # If no milestone is provided but date range is, try to find Chrome versions for that period
        if not milestone and start_date and end_date:
            try:
                # Get current version as a fallback
                current_version = self.get_current_chrome_version()
                
                # In a real implementation, you would query for Chrome versions released 
                # between start_date and end_date, and use those as milestones
                # For now, we'll just use the current version
                milestone = current_version
                
                print(f"Using Chrome version {milestone} for date range {start_date} to {end_date}")
            except Exception as e:
                print(f"Error finding Chrome versions for date range: {str(e)}")
        
        # Process all milestones if multiple versions were specified
        if isinstance(versions, list) and len(versions) > 1:
            results = []
            for version in versions:
                result = self._process_single_milestone(version, working_dir, category_filter, status_filter)
                results.append(result)
            return {
                "status": "success", 
                "message": f"Processed {len(results)} Chrome versions",
                "results": results
            }
        result = handler.process_chrome_features(working_dir=working_dir)
        # Process a single milestone
        return self._process_single_milestone(milestone, working_dir, category_filter, status_filter)
    
    def _process_single_milestone(self, milestone, working_dir, category_filter, status_filter):
        """Helper method to process a single Chrome milestone"""
        try:
            # If no milestone specified, get current stable version
            if milestone is None:
                milestone = self.get_current_chrome_version()
                if milestone is None:
                    return {
                        "status": "error",
                        "message": "Could not determine current Chrome version"
                    }
            
            # Build URL
            url = f"{self.base_url}?milestone={milestone}"
            print(f"Fetching features from URL: {url}")
            
            # Use crawl_website to get data
            response_text = crawl_website(url)
            
            # Clean response text
            cleaned_response = self._clean_json_response(response_text)
            
            # Parse JSON
            features_data = json.loads(cleaned_response)
            
            # Ensure features_data is a list
            if isinstance(features_data, dict):
                # Data might be under some key
                for key in ['features', 'data', 'items']:
                    if key in features_data and isinstance(features_data[key], list):
                        print(f"Found features array in key: '{key}'")
                        features_data = features_data[key]
                        break
                else:
                    # If still a dict, convert to list
                    print("Converting dict to list")
                    features_data = [features_data]
            
            # Filter features
            filtered_features = self._filter_features(features_data, category_filter, status_filter)
            
            # Process feature data
            processed_features = []
            for feature in filtered_features:
                processed_feature = {
                    "id": feature.get("id"),
                    "name": feature.get("name", "Unknown"),
                    "summary": feature.get("summary", ""),
                    "category": self._get_category_name(feature.get("category")),
                    "status": feature.get("impl_status_chrome", "Unknown"),
                    "feature_type": feature.get("feature_type", ""),
                    "created": self._format_date(feature.get("created")),
                    "updated": self._format_date(feature.get("updated")),
                    "spec_link": feature.get("spec_link", ""),
                    "bug_url": feature.get("bug_url", ""),
                    "blink_components": feature.get("blink_components", []),
                    "browsers": {
                        "chrome": feature.get("impl_status_chrome", ""),
                        "firefox": self._get_browser_status(feature.get("ff_views", {})),
                        "safari": self._get_browser_status(feature.get("safari_views", {})),
                        "edge": self._get_browser_status(feature.get("ie_views", {}))
                    }
                }
                processed_features.append(processed_feature)
            
            # # Save results
            # output_file = f"chrome_features_v{milestone}.json"
            # if working_dir:
            #     output_file = os.path.join(working_dir, output_file)
            
            # with open(output_file, 'w', encoding='utf-8') as f:
            #     json.dump({
            #         "milestone": milestone,
            #         "features_count": len(processed_features),
            #         "generated_at": datetime.now().isoformat(),
            #         "features": processed_features
            #     }, f, indent=2, ensure_ascii=False)
            
            # Generate summary
            summary = self._generate_summary(processed_features, milestone)
            
            return {
                "status": "success",
                "milestone": milestone,
                "features_found": len(processed_features),
                "output_file": os.path.abspath(output_file),
                "summary": summary
            }
            
        except Exception as e:
            return {
                "status": "error",
                "message": f"Error fetching Chrome features: {str(e)}"
            }
    
    def _filter_features(self, features: List[Dict], 
                        category_filter: Optional[List[str]] = None,
                        status_filter: Optional[List[str]] = None) -> List[Dict]:
        """Filter features based on category and status."""
        filtered = features
        
        # Filter by category
        if category_filter:
            category_ids = [self._get_category_id(cat) for cat in category_filter]
            filtered = [f for f in filtered if f.get("category") in category_ids]
        
        # Filter by status
        if status_filter:
            filtered = [f for f in filtered 
                       if any(status.lower() in f.get("impl_status_chrome", "").lower() 
                             for status in status_filter)]
        
        return filtered
    
    def _get_category_name(self, category_id: Any) -> str:
        """Convert category ID to name."""
        category_map = {
            1: "CSS",
            2: "HTML",
            3: "JavaScript",
            4: "DOM",
            5: "Network/Connectivity",
            6: "Security",
            7: "Storage",
            8: "Misc",
            9: "Web Components",
            10: "Input",
            11: "Performance",
            12: "WebRTC",
            13: "Graphics",
            14: "Audio/Video",
            15: "Rendering",
            16: "User Interface",
            17: "Media",
            18: "WebAssembly"
        }
        
        try:
            return category_map.get(int(category_id), "Other")
        except:
            return "Other"
    
    def _get_category_id(self, category_name: str) -> int:
        """Convert category name to ID."""
        category_map = {
            "css": 1,
            "html": 2,
            "javascript": 3,
            "dom": 4,
            "network": 5,
            "connectivity": 5,
            "security": 6,
            "storage": 7,
            "misc": 8,
            "web components": 9,
            "input": 10,
            "performance": 11,
            "webrtc": 12,
            "graphics": 13,
            "audio": 14,
            "video": 14,
            "rendering": 15,
            "user interface": 16,
            "ui": 16,
            "media": 17,
            "webassembly": 18,
            "wasm": 18
        }
        
        return category_map.get(category_name.lower(), 0)
    
    def _get_browser_status(self, browser_views: Dict) -> str:
        """Extract browser status from views object."""
        if isinstance(browser_views, dict):
            return browser_views.get("text", "Unknown")
        return "Unknown"
    
    def _format_date(self, date_str: str) -> str:
        """Format date string."""
        if not date_str:
            return ""
        
        try:
            # Parse ISO format date
            dt = datetime.fromisoformat(date_str.replace('Z', '+00:00'))
            return dt.strftime("%Y-%m-%d")
        except:
            return date_str
    
    def _generate_summary(self, features: List[Dict], milestone: int) -> Dict[str, Any]:
        """Generate a summary of the features."""
        # Count by status
        status_counts = {}
        for feature in features:
            status = feature["status"]
            status_counts[status] = status_counts.get(status, 0) + 1
        
        # Count by category
        category_counts = {}
        for feature in features:
            category = feature["category"]
            category_counts[category] = category_counts.get(category, 0) + 1
        
        # Get top features
        top_features = sorted(features, key=lambda x: x["updated"], reverse=True)[:5]
        
        return {
            "total_features": len(features),
            "status_breakdown": status_counts,
            "category_breakdown": category_counts,
            "top_features": [
                {
                    "name": f["name"],
                    "category": f["category"],
                    "status": f["status"]
                } for f in top_features
            ]
        }
    
    def _save_response_to_markdown(self, milestone: int, response_text: str, cleaned_json: str, working_dir: str = "") -> str:
        """
        Save the raw response and cleaned JSON to a Markdown file for inspection.
        
        Args:
            milestone: Chrome version number
            response_text: Raw response text from the API
            cleaned_json: Cleaned JSON text
            working_dir: Directory to save the output file
            
        Returns:
            str: Path to the saved Markdown file
        """
        filename = f"chrome_features_v{milestone}_response.md"
        if working_dir:
            filename = os.path.join(working_dir, filename)
        
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(f"# Chrome Features API Response - v{milestone}\n\n")
            
            f.write("## API URL\n")
            f.write(f"```\n{self.base_url}?milestone={milestone}\n```\n\n")
            
            # Check if response contains ")]}'" prefix
            has_prefix = False
            prefix_type = None
            if response_text.startswith(")]}'\n"):
                has_prefix = True
                prefix_type = ")]}'\n"
            elif response_text.startswith(")]}'"):
                has_prefix = True
                prefix_type = ")]}''"
            
            if has_prefix:
                f.write(f"**Note:** Response contains `{prefix_type}` prefix, which needs to be removed before parsing JSON.\n\n")
            
            f.write("## Raw Response Preview\n")
            f.write("```\n")
            # Limit the length of raw response to avoid large file size
            f.write(response_text[:5000])
            if len(response_text) > 5000:
                f.write("\n... (response truncated) ...\n")
            f.write("\n```\n\n")
            
            f.write("## Cleaned JSON\n")
            f.write("```json\n")
            f.write(cleaned_json)
            f.write("\n```\n\n")
            
            # Add some parsing info
            try:
                json_data = json.loads(cleaned_json)
                f.write("## JSON Structure Info\n")
                
                if isinstance(json_data, dict):
                    f.write(f"- Type: Dictionary\n")
                    f.write(f"- Keys: {list(json_data.keys())}\n")
                    
                    # Check for nested feature lists
                    for key in ['features', 'data', 'items']:
                        if key in json_data and isinstance(json_data[key], list):
                            f.write(f"- Found features list in key: '{key}' with {len(json_data[key])} items\n")
                            if json_data[key]:
                                f.write(f"- First item keys: {list(json_data[key][0].keys())}\n")
                
                elif isinstance(json_data, list):
                    f.write(f"- Type: List\n")
                    f.write(f"- Length: {len(json_data)}\n")
                    if json_data:
                        f.write(f"- First item type: {type(json_data[0])}\n")
                        if isinstance(json_data[0], dict):
                            f.write(f"- First item keys: {list(json_data[0].keys())}\n")
            except Exception as e:
                f.write(f"Error analyzing JSON: {str(e)}\n")
        
        return os.path.abspath(filename)
    
    def save_features_to_markdown(self, features_data: List[Dict], milestone: int, working_dir: str = "") -> str:
        """
        Save Chrome features in the exact format that matches the Chrome Status website.
        
        Args:
            features_data: List of feature dictionaries
            milestone: Chrome version number
            working_dir: Directory to save the output file
            
        Returns:
            str: Path to the saved Markdown file
        """
        filename = f"chrome_features_v{milestone}_summary.md"
        if working_dir:
            filename = os.path.join(working_dir, filename)
        
        # Group features by status
        enabled_features = []
        dev_trial_features = []
        origin_trial_features = []
        stepped_rollout_features = []
        
        for feature in features_data:
            name = feature.get("name", "Unknown")
            status = feature.get("impl_status_chrome", "").lower()
            
            if "enabled by default" in status or "shipped" in status:
                enabled_features.append(name)
            elif "developer trial" in status or "behind a flag" in status:
                dev_trial_features.append(name)
            elif "origin trial" in status:
                origin_trial_features.append(name)
            elif "rollout" in status:
                stepped_rollout_features.append(name)
        
        # Get the current date for the "X weeks ago" text
        # For version 137 we use the exact date from the provided example
        release_date = "May 21, 2025"
        beta_period = "Apr 30 - May 15"
        
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(f"<!-- filepath: {filename} -->\n")
            f.write(f"# Chrome {milestone}\n\n")
            f.write(f"Beta was {beta_period}\n")
            f.write(f"Stable 4 weeks ago ({release_date})\n\n")
            
            f.write("## Features in this release:\n\n")
            
            if enabled_features:
                f.write("### Enabled by default\n")
                for feature in sorted(enabled_features):
                    f.write(f"- {feature}\n")
                f.write("\n")
            
            if dev_trial_features:
                f.write("### In developer trial (Behind a flag)\n")
                for feature in sorted(dev_trial_features):
                    f.write(f"- {feature}\n")
                f.write("\n")
            
            if origin_trial_features:
                f.write("### Origin trial\n")
                for feature in sorted(origin_trial_features):
                    f.write(f"- {feature}\n")
                f.write("\n")
            
            if stepped_rollout_features:
                f.write("### Stepped rollout\n")
                for feature in sorted(stepped_rollout_features):
                    f.write(f"- {feature}\n")
        
        return os.path.abspath(filename)
    
    def save_raw_response_to_markdown(self, response_text: str, milestone: int, working_dir: str = "") -> str:
        """
        This method is kept for backward compatibility but will be deprecated.
        Directly use parse_features_to_markdown instead.
        """
        print("Warning: save_raw_response_to_markdown is deprecated. Use parse_features_to_markdown instead.")
        return self.parse_features_to_markdown(response_text, milestone, working_dir)
    
    def parse_features_to_markdown(self, response_text: str, milestone: int, working_dir: str = "") -> str:
        """
        Parse Chrome features API response and save as a structured Markdown file.
        
        Args:
            response_text: Raw response text from the API
            milestone: Chrome version number
            working_dir: Directory to save the output file
            
        Returns:
            str: Path to the saved Markdown file
        """
        filename = f"chrome_features_v{milestone}.md"
        if working_dir:
            filename = os.path.join(working_dir, filename)
        
        # Clean the response text - remove ")]}'" prefix
        cleaned_text = response_text
        if cleaned_text.startswith(")]}'\n"):
            cleaned_text = cleaned_text[5:]
        elif cleaned_text.startswith(")]}'"):
            cleaned_text = cleaned_text[4:]
        
        try:
            json_data = json.loads(cleaned_text)
            features_by_type = json_data.get("features_by_type", {})
            total_count = json_data.get("total_count", 0)
            
            # Generate a summary of the features
            summary = self._generate_version_summary(features_by_type)
            
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(f"<!-- filepath: {filename} -->\n")
                f.write(f"# Chrome {milestone} Features\n\n")
                f.write(f"Total features: {total_count}\n\n")
                
                # Add the summary section - changed to English
                f.write("## Version Update Summary\n\n")
                f.write(summary)
                f.write("\n\n")
                
                # Remove the single index table and instead place tables after each category
                
                # Continue with the existing feature details
                for feature_type, features in features_by_type.items():
                    if not features:
                        continue
                    
                    f.write(f"## {feature_type} ({len(features)})\n\n")
                    
                    # Add feature table for this category
                    f.write("| Feature | ID | Status |\n")
                    f.write("|---------|----|---------|\n")
                    
                    for feature in features:
                        name = feature.get("name", "Unnamed Feature")
                        feature_id = feature.get("id", "Unknown ID")
                        is_released = feature.get("is_released", False)
                        status_text = "Released" if is_released else "Not Released"
                        
                        # Create a link-friendly anchor
                        link_name = name.lower().replace(' ', '-')
                        link_name = re.sub(r'[^\w\-]', '', link_name)
                        
                        # Add row to the table
                        f.write(f"| [{name}](#{link_name}) | {feature_id} | {status_text} |\n")
                    
                    f.write("\n\n")
                    
                    # Now output the detailed feature descriptions
                    for feature in features:
                        name = feature.get("name", "Unnamed Feature")
                        feature_id = feature.get("id", "Unknown ID")
                        summary = feature.get("summary", "No summary provided")
                        is_released = feature.get("is_released", False)
                        status_text = "Released" if is_released else "Not Released"
                        
                        f.write(f"### {name}\n\n")
                        f.write(f"**ID:** {feature_id}  \n")
                        f.write(f"**Status:** {status_text}  \n\n")
                        f.write(f"{summary}\n\n")
                        
                        # Add browser compatibility info
                        browsers = feature.get("browsers", {})
                        if browsers:
                            f.write("#### Browser Compatibility\n\n")
                            
                            chrome_info = browsers.get("chrome", {})
                            chrome_status = chrome_info.get("status", {}).get("text", "Unknown")
                            f.write(f"- **Chrome:** {chrome_status}\n")
                            
                            ff_info = browsers.get("ff", {}).get("view", {})
                            ff_status = ff_info.get("text", "No signal")
                            f.write(f"- **Firefox:** {ff_status}\n")
                            
                            safari_info = browsers.get("safari", {}).get("view", {})
                            safari_status = safari_info.get("text", "No signal")
                            f.write(f"- **Safari:** {safari_status}\n\n")
                        
                        # Add standards info
                        standards = feature.get("standards", {})
                        if standards and standards.get("spec"):
                            f.write("#### Standards\n\n")
                            f.write(f"- **Spec:** [{standards.get('spec')}]({standards.get('spec')})\n")
                            maturity = standards.get("maturity", {})
                            if maturity and maturity.get("text"):
                                f.write(f"- **Maturity:** {maturity.get('text')}\n\n")
                        
                        f.write("---\n\n")
                
                return os.path.abspath(filename)
                
        except Exception as e:
            print(f"JSON parsing failed: {str(e)}")
            # Fall back to saving raw response
            return self.save_raw_response_to_markdown(response_text, milestone, working_dir)

    def _generate_version_summary(self, features_by_type: Dict[str, List[Dict]]) -> str:
        """
        Generate a concise summary of the Chrome version features.
        
        Args:
            features_by_type: Dictionary of features grouped by type
            
        Returns:
            str: Markdown formatted summary
        """
        summary = []
        
        # Count features by status
        status_counts = {
            "Released": 0,
            "Not Released": 0
        }
        
        # Count features by category and collect key highlights
        categories = {}
        highlights = []
        
        # Process all features
        for feature_type, features in features_by_type.items():
            feature_count = len(features)
            summary.append(f"- **{feature_type}**: {feature_count} features")
            
            for feature in features:
                # Count by status
                is_released = feature.get("is_released", False)
                status = "Released" if is_released else "Not Released"
                status_counts[status] += 1
                
                # Extract browser compatibility info
                browsers = feature.get("browsers", {})
                chrome_info = browsers.get("chrome", {})
                chrome_status = chrome_info.get("status", {}).get("text", "")
                
                # Add to highlights if it's a significant feature
                if chrome_status in ["Shipped/Shipping", "In developer trial", "Origin trial"]:
                    name = feature.get("name", "")
                    if name and len(name) < 100:  # Avoid very long names
                        highlights.append
                
                # Categorize by standards area
                standards = feature.get("standards", {})
                if standards and standards.get("spec"):
                    spec_url = standards.get("spec", "")
                    category = "Other"
                    
                    # Try to determine category from spec URL
                    if "css" in spec_url.lower():
                        category = "CSS"
                    elif "html" in spec_url.lower():
                        category = "HTML"
                    elif "javascript" in spec_url.lower() or "ecmascript" in spec_url.lower():
                        category = "JavaScript"
                    elif "webassembly" in spec_url.lower() or "wasm" in spec_url.lower():
                        category = "WebAssembly"
                    elif "security" in spec_url.lower():
                        category = "Security"
                    elif "gpuweb" in spec_url.lower() or "webgpu" in spec_url.lower():
                        category = "WebGPU"
                    
                    categories[category] = categories.get(category, 0) + 1
        
        # Format the summary - changed to English
        summary_text = "This Chrome version includes the following updates:\n\n"
        
        # Add status counts
        summary_text += f"- Released features: {status_counts['Released']}\n"
        summary_text += f"- In-development features: {status_counts['Not Released']}\n\n"
        
        # Add breakdown by feature type
        summary_text += "Categorized by feature type:\n"
        summary_text += "\n".join(summary)
        summary_text += "\n\n"
        
        # Add breakdown by standards area if available
        if categories:
            summary_text += "Categorized by standards area:\n"
            for category, count in sorted(categories.items(), key=lambda x: x[1], reverse=True):
                summary_text += f"- {category}: {count}\n"
            summary_text += "\n"
        
        # Add key highlights (limit to 5)
        if highlights:
            summary_text += "Key feature highlights:\n"
            for highlight in highlights[:5]:
                summary_text += f"- {highlight}\n"
        
        return summary_text
    
    def process_chrome_features(self, milestone: Optional[int] = None, working_dir: str = "") -> Dict[str, Any]:
        """
        Process Chrome features for a specific milestone or the current version.
        
        Args:
            milestone: Chrome version milestone number. If None, uses current stable version.
            working_dir: Directory to save output files
            
        Returns:
            Dict: Result containing status, version info, and output file path
        """
        try:
            # If no milestone specified, get current stable version
            if milestone is None:
                milestone = self.get_current_chrome_version()
                if milestone is None:
                    print("Could not determine current Chrome version, using a default version")
                    milestone = 138  # Default version
            
            # Build URL
            url = f"{self.base_url}?milestone={milestone}"
            print(f"📌 Using URL: {url}")
            
            # Get API response
            print(f"Fetching API response...")
            raw_response = crawl_website(url)
            
            # Parse to Markdown
            parsed_file = self.parse_features_to_markdown(
                raw_response,
                milestone,
                working_dir
            )
            
            print(f"\n✅ Successfully parsed Chrome {milestone} features")
            print(f"📄 Feature file saved to: {parsed_file}")
            
            return {
                "status": "success",
                "milestone": milestone,
                "output_file": parsed_file
            }
            
        except Exception as e:
            print(f"Error fetching or parsing response data: {str(e)}")
            import traceback
            traceback.print_exc()
            
            return {
                "status": "error",
                "message": str(e)
            }

# Example usage
if __name__ == "__main__":
    # Create handler instance
    handler = ChromeFeaturesHandler()
    print("Getting current Chrome version...")
    
    # Use working directory from environment variable
    working_dir = os.environ.get('WORKING_DIR', '.')
    
    # Call fetch_data instead of process_chrome_features
    result = handler.fetch_data(
        product_name="Chrome",
        working_dir=working_dir
    )
    
    # if result["status"] == "error":
    #     print(f"Processing failed: {result.get('message', 'Unknown error')}")
    # else:
    #     print(f"Successfully processed Chrome features")
    #     print(f"Features found: {result.get('features_found', 0)}")
    #     print(f"Output file: {result.get('output_file', '')}")


import os
import csv
from typing import List
import re
import glob
import shutil
# Optional import for markitdown - will also be imported locally when needed
try:
    from markitdown import MarkItDown
except ImportError:
    # Module will be imported when needed in the function
    pass
# Stop words

class FileUtils:
    @staticmethod
    def replace_stop_words(text: str) -> str:
        """Replace stop words with asterisks."""
        STOP_WORDS = ['fuck', 'shit', 'damn', 'ass', 'bitch', 'crap']

        if not text:
            return text
        
        result = text
        for word in STOP_WORDS:
            pattern = r'\b' + word + r'\b'
            result = re.sub(pattern, '***', result, flags=re.IGNORECASE)
        
        return result

    @staticmethod
    def generate_markdown_tool(content: str, filename: str, working_path: str):
        """
        Write the given content to a Markdown file in the working directory with a customizable filename.
                
        Args:
            content (str): The content to write to the Markdown file.
            filename (str, optional): The name of the output file.
                
        Returns:
            str: Path to the saved Markdown file.
        """
        try:
            # Ensure the filename has .md extension
            if not filename.endswith('.md'):
                filename = f"{filename}.md"
                
            # Define the output file path
            output_file = os.path.join(working_path, filename)
                    
            # Write content to the file
            with open(output_file, 'w', encoding='utf-8') as file:
                file.write(content)
                    
            return f"Content successfully written to {output_file}"
        except Exception as e:
            return f"Error writing to markdown file: {str(e)}"
        
    @staticmethod
    def upload_files_to_working_dir_tool(file_paths: List[str], working_path): #handle the file uploaded from the client
        """
        Copies the CSV files uploaded from the client chat to the working directory and renames them with an "_intermediate" suffix.

        Args:
            file_paths (List[str]): List of file paths uploaded from the client chat.
        """
        copied_files = []
        for file_path in file_paths:
            try:
                # Ensure the file is a CSV file
                if not file_path.lower().endswith('.csv'):
                    return f"Error: Only CSV files are allowed. Invalid file: {file_path}"

                # Add "_intermediate" suffix to the file name
                base_name = os.path.splitext(os.path.basename(file_path))[0]
                destination = os.path.join(working_path, f"{base_name}_intermediate.csv")

                # Copy the file to the working directory
                shutil.copy(file_path, destination)
                copied_files.append(destination)
            except Exception as e:
                return f"Error copying file {file_path}: {str(e)}"

        return f"Copied {len(copied_files)} files to working directory: {', '.join(copied_files)}"
    
    @staticmethod
    def convert_to_markdown_tool(file_path: str, working_path: str):
        """
        Convert a document (doc/excel/ppt/pdf/images/csv/json/xml) to markdown format using MarkItDown.

        Args:
            file_path (str): Path to the input document file
            working_path (str): Path to the working directory to save output

        Returns:
            str: Path to the generated markdown file or error message
        """
        try:
            # Import here to avoid circular imports
            from markitdown import MarkItDown
            
            md = MarkItDown()
            result = md.convert(file_path)
            
            # Generate output filename
            base_name = os.path.splitext(os.path.basename(file_path))[0]
            output_file = os.path.join(working_path, f"{base_name}.md")
            
            # Write the converted content to file
            with open(output_file, 'w', encoding='utf-8') as f:
                f.write(result.text_content)
                
            return f"File converted successfully: {output_file}"
        except Exception as e:
            return f"Error converting file: {str(e)}"


"""
Metadata for KPI_BrowserMinutes_All table
"""

DESCRIPTION = """
Browser Minutes Share on Windows
"""

SAMPLE = \
"""
Date,BrowserName,CustomerGroup,CustomerSubGroup,EducationSegment,Manufacturer,OsSku,Os,FormFactorFamily,FormFactor,A14Region,DeviceType,NewDeviceFlag,HasMSA,DefaultInternetBrowser,TpidSegmentGroup,TpidSegment,TpidSubSegment,TpidIndustry,TpidIndustrySummary,TpidIsS500,TpidIsS2500,TelemetryActiveDevices_Daily,TelemetryBrowsingMinutes_Daily,TelemetryPrimaryActiveDevices_Daily,TelemetryPrimaryBrowsingMinutes_Daily,ActiveDevices_Daily,BrowsingMinutes_Daily,PrimaryActiveDevices_Daily,PrimaryBrowsingMinutes_Daily,TelemetryActiveDevices_RL7,TelemetryBrowsingMinutes_RL7,TelemetryPrimaryActiveDevices_RL7,TelemetryPrimaryBrowsingMinutes_RL7,ActiveDevices_RL7,BrowsingMinutes_RL7,PrimaryActiveDevices_RL7,PrimaryBrowsingMinutes_RL7,TelemetryActiveDevices_RL28,TelemetryBrowsingMinutes_RL28,TelemetryPrimaryActiveDevices_RL28,TelemetryPrimaryBrowsingMinutes_RL28,ActiveDevices_RL28,BrowsingMinutes_RL28,PrimaryActiveDevices_RL28,PrimaryBrowsingMinutes_RL28
2021-09-15 00:00:00,2345 Explorer,Commercial,Commercial Enterprise,Education HED,Other,Core,Windows 10 21H1 GA,PC,Convertible,ANZ,New,-1,Yes,2345Explorer,EDU,Major Public Sector,Major - Education,Higher Education,Education,0,0,0,0,0,0,0,0,0,0,0,0.0,0,0,0.0,0.0,0,0,1,11.617733333333332,0,0,1.8905642327532124,21.964071105664754,0,0
2021-09-15 00:00:00,2345 Explorer,Commercial,Commercial Enterprise,Education HED,Other,Core,Windows 10 21H1 GA,PC,Desktop,ANZ,Upgrade,-1,Yes,Chrome,EDU,Major Public Sector,Major - Education,Higher Education,Education,0,0,0,0,0,0,0,0,0,0,1,0.08125,0,0,1.8854546685368758,0.15319319181862118,0,0,1,1.5554833333333333,0,0,1.8854546685368758,2.9327933126646344,0,0
2021-09-15 00:00:00,2345 Explorer,Commercial,Commercial Enterprise,Education HED,Other,Core,Windows 10 19H2 GA,PC,Notebook,ANZ,New,-1,No,2345Explorer,EDU,Major Public Sector,Major - Education,Higher Education,Education,0,0,0,0,0,0,0,0,0,0,0,0.0,0,0,0.0,0.0,0,0,1,6.533833333333333,0,0,8.235838360017018,53.811595204624524,0,0
"""

FILTER_COLUMNS = {
    "A14Region": {
        "description": "User demographic region",
        "values": ["Greater China", "United States", "Latam", "APAC", "MEA", "Japan", "India", "UK", "Korea", "Canada", "ANZ", "Unknown", ""]
    },
    "start_date": {
        "description": "Start date for data analysis",
        "example": "2025-04-15"
    },
    "end_date": {
        "description": "End date for data analysis",
        "example": "2025-05-15"
    },
    "segment": {
        "description": "Customer segment for data analysis",
        "values": ["Consumer", "Commercial", "Education", "Unknown"]
    },
}

SQL_TEMPLATES = [
    {
        "name": "edge_browser_minutes_windows_R7",
        "description": "Edge browser minutes for Windows,average over 7 days",
        "template": """
SELECT toStartOfDay(toDateTime(`Date`)) AS `__timestamp`,
       sumIf(BrowsingMinutes_RL7, BrowserName in('Ms1p', 'Edge (Overall) + IE')) / 
       sumIf(BrowsingMinutes_RL7, BrowserName in('AllUp', 'All Browsers')) AS `Total_MSFT_BSOM_R7`
FROM KPI_BrowserMinutes_All
WHERE `Date` >= toDateTime('{start_date}')
  AND `Date` < toDateTime('{end_date}')
  AND ((Date>=date_add(DAY, -720, now()))
       AND (FormFactorFamily IN ('PC',
                                 'Tablet',
                                 'Other',
                                 'Unknown'))
       AND (BrowserName IN ('AllUp',
                            'All Browsers',
                            'Ms1p',
                            'Edge (Overall) + IE',
                            'Edge (Overall)',
                            'Edge'))
       AND (DayOfWeek(Date) IN ('3'))
       AND (A14Region NOT IN ('CEE',
                              'France',
                              'Germany',
                              'Western Europe'))
       {region_filter}
       {segment_filter}
GROUP BY toStartOfDay(toDateTime(`Date`))
ORDER BY toStartOfDay(toDateTime(`Date`)) ASC
LIMIT 50000;
""",
        "required_filters": ["start_date", "end_date"],
        "optional_filters": ["A14Region", "segment"],
        "template_params": {
            "region_filter": lambda params: f"AND (A14Region IN ('{params.get('A14Region')}'))" if params.get('A14Region') and params.get('A14Region') != '' else "",
            "segment_filter": lambda params: (
                "AND (IF(CustomerGroup == 'Consumer', 'Consumer', "
                "IF(CustomerGroup == 'Commercial', IF(startsWith(EducationSegment, 'Education'), 'Education', 'Commercial'), 'Unknown')) "
                f"IN ('{params.get('segment', '')}'))" if params.get('segment') else ""
            )
        }
    }
]



"""Exception classes for GitHub Pages publishing."""


class PublishError(Exception):
    """Base exception for publishing errors."""
    pass


class GitOperationError(PublishError):
    """Raised when a Git operation fails."""
    pass


class FileProcessingError(PublishError):
    """Raised when file processing fails."""
    pass


class ValidationError(PublishError):
    """Raised when validation fails."""
    pass


class UncommittedChangesError(PublishError):
    """Raised when there are uncommitted changes in the repository."""
    pass



import os
from typing import Optional
"""
Constants used in the PM Studio MCP server.
"""

# =====================
# Greeting message template
# =====================
GREETING_TEMPLATE = "hello, {name}! How can I help you today? I can help you to do competitor analysis, user feedback summary, write docs and more!"

# =====================
# Configuration constants (placeholders, please update with real values as needed)
# Warning: Do not hardcode sensitive information in production code, you should consider adding environment variables in .env file.
# =====================

# Microsoft tenant ID for authentication
MICROSOFT_TENANT_ID = "72f988bf-86f1-41af-91ab-2d7cd011db47"  # Microsoft tenant ID

# Titan API configuration
TITAN_CLIENT_ID = "dcca0492-ea09-452c-bf98-3750d4331d33"  # Titan API client ID
TITAN_ENDPOINT = "https://titanapi.westus2.cloudapp.azure.com/v2/query"  # Titan API endpoint
TITAN_SCOPE = "api://dcca0492-ea09-452c-bf98-3750d4331d33/signin"  # Titan API scope

# =====================
# Graph API Team Configuration
# =====================
# Microsoft Graph Client IDs mapped to team aliases
GRAPH_TEAM_MAPPING = {
    # Edge Consumer Team
    "0facc9cb-0dc6-4c59-8306-f9b6c154ca19": [
        "yche",
        "gajie", 
        "juanliu",
        "mile",
        "yancheng",
        "dingxiao"
    ],
    # Edge Mobile Team  
    "c9ec9ed4-22fc-492d-be60-a1527ee6ffe2": [
        "emilywu",
        "hongjunqiu",
        "yingzhuang", 
        "lmike",
        "yazhouzhou",
        "shengjieshi",
        "v-xiaomengli",
        "wenyuansu",
        "jinghuama"
    ],
    # SA Bill Team
    "b2137a41-1db3-47a1-84fb-4893d018771b": [
        "tajie",
        "xiaoxch",
        "chenxitan",
        "carmenwei",
        "liyayong",
        "yongweizhang"
    ],
    # SA Kelly Team
    "45056f4f-9397-47b6-a0e5-bd7ce7e6165c": [
        "yugon",
        "danliu",
        "eviema",
        "angliu",
        "menghuihu",
        "emmaxu",
        "zhangjingwei"
    ],
    # WebComFun Team
    "f008f6ae-1563-4bcf-88ff-15d34cfbe60d": [
        "michachen",
        "chfen",
        "alexyuan",
        "lingyanzhao",
        "nanyin",
        "siyangliu"
    ]
}

# Reddit API configuration
# Reddit API client ID
REDDIT_CLIENT_ID = ""
# Reddit API client secret
REDDIT_CLIENT_SECRET = ""

# Data.ai API configuration
# Data.ai API key
DATA_AI_API_KEY = ""

# Unwrap API configuration
# Unwrap API access token
UNWRAP_ACCESS_TOKEN = ""  

# Path of your working directory, this is used to store output files
WORKING_PATH = ""

# GitHub Copilot token, should be set as an environment variable for security
GITHUB_TOKEN = ""  

def get_user_graph_id() -> str:
    """
    Auto-detect Graph Client ID by mapping user alias to team configuration.
    
    Returns:
        str: Graph Client ID for the user's team, or empty string if not found
    """
    # Use UserUtils to get current user alias and map to team
    try:
        from pm_studio_mcp.utils.graph.user import UserUtils
        
        user_alias = UserUtils.get_current_user_alias()
        if user_alias:
            user_alias = user_alias.strip().lower()
            print(f"Retrieved user alias: '{user_alias}'", flush=True)
            
            # Look up client ID from team mapping
            for graph_client_id, aliases in GRAPH_TEAM_MAPPING.items():
                if user_alias in aliases:
                    print(f"Found Graph Client ID: {graph_client_id[:8]}...", flush=True)
                    return graph_client_id
            
            print(f"User alias '{user_alias}' not found in team mapping", flush=True)
            
    except Exception as e:
        print(f"UserUtils failed: {e}", flush=True)
    
    print("No Graph Client ID found", flush=True)
    return ""

class Config:
    _instance = None
    _initialized = False

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        if not self._initialized:
            self._initialized = True
            self._initialize_config()

    def _initialize_config(self):

        # Greeting message configuration
        self.GREETING_TEMPLATE = os.environ.get('GREETING_TEMPLATE', GREETING_TEMPLATE)

        # Working directory configuration
        self.WORKING_PATH = os.environ.get('WORKING_PATH', os.path.join(os.path.dirname(os.path.abspath(__file__)), '../../temp'))
        os.makedirs(self.WORKING_PATH, exist_ok=True)

        # Reddit API configuration
        self.REDDIT_CLIENT_ID = self._get_config_value('REDDIT_CLIENT_ID', REDDIT_CLIENT_ID)
        self.REDDIT_CLIENT_SECRET = self._get_config_value('REDDIT_CLIENT_SECRET', REDDIT_CLIENT_SECRET)

        # Data.ai API configuration
        self.DATA_AI_API_KEY = self._get_config_value('DATA_AI_API_KEY', DATA_AI_API_KEY)

        # Microsoft Graph API configuration
        self.GRAPH_CLIENT_ID = self._get_config_value('GRAPH_CLIENT_ID', get_user_graph_id()) 

        # Microsoft authentication configuration
        self.MICROSOFT_TENANT_ID = self._get_config_value('MICROSOFT_TENANT_ID', MICROSOFT_TENANT_ID)

        # Unwrap API configuration
        self.UNWRAP_ACCESS_TOKEN = self._get_config_value('UNWRAP_ACCESS_TOKEN', UNWRAP_ACCESS_TOKEN) 

        # Titan API configuration
        self.TITAN_CLIENT_ID = self._get_config_value('TITAN_CLIENT_ID', TITAN_CLIENT_ID)
        self.TITAN_ENDPOINT = self._get_config_value('TITAN_ENDPOINT',  TITAN_ENDPOINT)
        self.TITAN_SCOPE = self._get_config_value('TITAN_SCOPE', TITAN_SCOPE)

        # Google Ads API configuration
        self.GOOGLE_ADS_DEVELOPER_TOKEN = self._get_config_value('GOOGLE_ADS_DEVELOPER_TOKEN', '')
        self.GOOGLE_ADS_LOGIN_CUSTOMER_ID = self._get_config_value('GOOGLE_ADS_LOGIN_CUSTOMER_ID', '')
        self.GOOGLE_ADS_CLIENT_SECRET_JSON = self._get_config_value('GOOGLE_ADS_CLIENT_SECRET_JSON', '')
        self.GOOGLE_ADS_CREDENTIALS_JSON = self._get_config_value('GOOGLE_ADS_CREDENTIALS_JSON', '')

        # GitHub Copilot configuration
        # Note: GitHub Copilot token should be set as an environment variable for security
        self.GITHUB_TOKEN = self._get_config_value('GITHUB_TOKEN', GITHUB_TOKEN)

    def _get_config_value(self, env_var: str, default_value: str) -> str:
        value = os.environ.get(env_var)
        if value:
            print(f"Using {env_var} from environment")
            return value
        print(f"Using default value for {env_var}")
        return default_value

# Create global config instance
config = Config()


import requests
import markdown
import base64
import re
from .auth import AuthUtils
from .user import UserUtils

class ChatUtils:
    """Utilities for accessing Microsoft Teams chat data via MS Graph API"""

    @staticmethod
    def append_signature(html: str) -> str:
        signature = "<br><br><p style=\"color: grey;\"> Sent via <a href='https://aka.ms/pmstudio'>PM Studio</a></p>"
        return html + signature

    @staticmethod
    def get_chat_members(chat_id: str):
        access_token = AuthUtils().login()
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json"
        }
        members_url = f"https://graph.microsoft.com/v1.0/chats/{chat_id}/members"
        resp = requests.get(members_url, headers=headers)
        if resp.status_code == 200:
            return resp.json().get("value", [])
        else:
            print(f"Error retrieving chat members: {resp.text}")
            return None

    @staticmethod
    def get_group_chat_id_by_name(topic: str):
        print("Retrieving group chat ID...")
        access_token = AuthUtils().login()
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json"
        }
        params = {"$top": 50}
        chats_url = "https://graph.microsoft.com/v1.0/me/chats"

        while chats_url:
            chats_response = requests.get(chats_url, headers=headers, params=params)
            if chats_response.status_code != 200:
                print(f"Error retrieving chats: {chats_response.text}")
                return False

            for chat in chats_response.json().get("value", []):
                if chat.get("topic") and topic in chat.get("topic", ""):
                    print(chat.get("id"))
                    return chat.get("id")

            chats_url = chats_response.json().get("@odata.nextLink")
            params = None

        print(f"Chat with topic '{topic}' not found.")
        return None

    @staticmethod
    def match_member(mention_lower: str, member: dict) -> bool:
        user = member.get("user", {})
        email = member.get("email", "") or user.get("email", "")
        display = member.get("displayName", "") or user.get("displayName", "")
        alias = email.split("@")[0] if email else ""

        email = email.lower()
        display = display.lower()
        alias = alias.lower()
        display_no_space = display.replace(" ", "").lower()

        return (
            mention_lower == email or
            mention_lower == display or
            mention_lower == alias or
            alias.startswith(mention_lower) or
            mention_lower in display or
            mention_lower == display_no_space
        )

    @staticmethod
    def prepare_message_with_mentions(message: str, chat_id: str):
        mention_pattern = r'@([\w.\-]+)'
        mentions_found = re.findall(mention_pattern, message)

        if not mentions_found:
            return {
                "body": {
                    "contentType": "html",
                    "content": ChatUtils.append_signature(message.replace('\n', '<br>'))
                }
            }

        members = ChatUtils.get_chat_members(chat_id)
        if members is None:
            return {
                "body": {
                    "contentType": "html",
                    "content": ChatUtils.append_signature(message.replace('\n', '<br>'))
                }
            }

        html_content = message
        mentions_array = []

        for mention_text in mentions_found:
            mention_lower = mention_text.lower()
            matched_member = None

            for member in members:
                if ChatUtils.match_member(mention_lower, member):
                    matched_member = member
                    break

            if matched_member:
                user_obj = matched_member.get("user", {})
                user_id = user_obj.get("id") or matched_member.get("userId")
                display_name = matched_member.get("displayName") or user_obj.get("displayName") or mention_text

                if not user_id:
                    continue

                mention_id = len(mentions_array)
                mentions_array.append({
                    "id": mention_id,
                    "mentionText": f"@{display_name}",
                    "mentioned": {
                        "@odata.type": "#microsoft.graph.chatMessageMentionedIdentitySet",
                        "user": {
                            "id": user_id,
                            "displayName": display_name
                        }
                    }
                })

                html_content = ChatUtils._replace_first_mention(html_content, mention_text, display_name, mention_id)
            else:
                continue

        html_content = html_content.replace('\n', '<br>')
        html_content = ChatUtils.append_signature(html_content)

        return {
            "body": {
                "contentType": "html",
                "content": html_content
            },
            "mentions": mentions_array
        }

    @staticmethod
    def _replace_first_mention(text: str, mention_text: str, display_name: str, mention_id: int) -> str:
        pattern = f'@{re.escape(mention_text)}'
        replacement = f'<at id="{mention_id}">@{display_name}</at>'
        match = re.search(pattern, text)
        if not match:
            return text
        start, end = match.span()
        return text[:start] + replacement + text[end:]

    @staticmethod
    def send_chat_with_id(chat_id: str, message: str, image_path: str = None):
        print("Sending message to chat...")
        print("imgpath = ", image_path, flush=True)

        if image_path:  # image_path is not empty 
            # Read image as binary and convert to base64 string
            with open(image_path, "rb") as image_file:
                image_base64 = base64.b64encode(image_file.read()).decode('utf-8')

            # Prepare temporaryId reference
            temporary_id = "1"

            # Format message with image
            image_html = f'<div>{message}</div><img src="../hostedContents/{temporary_id}/$value" height="297" width="297">'
            message = f"<br/>{image_html}<br/>"


        access_token = AuthUtils().login()
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json"
        }

        if chat_id == "self":
            print("Sending note to myself...")
            endpoint = "https://graph.microsoft.com/v1.0/me/chats/48:notes/messages"
            # If sending to self, we don't need mentions or hosted contents
            payload = {
                        "body": {
                            "contentType": "html",
                            "content": ChatUtils.append_signature(message)
                        }
            }

        else:
            print(f"Sending note to group chat: {chat_id}")
            endpoint = f"https://graph.microsoft.com/v1.0/me/chats/{chat_id}/messages"
            payload = ChatUtils.prepare_message_with_mentions(message, chat_id)


        if image_path: 
            # Read image as binary and convert to base64 string
            payload["hostedContents"] = [
                {
                    "@microsoft.graph.temporaryId": temporary_id,
                    "contentBytes": image_base64,
                    "contentType": "image/png"  # Ensure the content type is specified
                }
            ]


        try:
            response = requests.post(url=endpoint, headers=headers, json=payload)
            print(f"Response status code: {response.status_code}")

            if response is None:
                return {
                    "status": "error",
                    "message": "No response received from the API."
                }

            if response.status_code == 201:
                print("Note sent successfully.")
                return {
                    "status": "success",
                    "message": "Note sent successfully."
                }
            else:
                return {
                    "status": "error",
                    "message": f"Error sending note: {response.text}"
                }

        except requests.exceptions.RequestException as e:
            return {
                "status": "error",
                "message": f"Error sending note: {str(e)}"
            }

    @staticmethod
    def send_message_to_chat(chat_type: str, chat_name: str, message: str, user_index: int = None , image_path: str = None):
        print("Sending note in Microsoft Teams...")
        messageHTML = markdown.markdown(message)
       
        if chat_type == "myself":
            return ChatUtils.send_chat_with_id("self", messageHTML,image_path)
        elif chat_type == "group":
            chat_id = ChatUtils.get_group_chat_id_by_name(chat_name)
            if chat_id is None:
                return {"status": "error", "message": f"Chat with topic '{chat_name}' not found."}
            return ChatUtils.send_chat_with_id(chat_id, messageHTML, image_path)
        elif chat_type == "person":
            # First try to find an existing one-on-one chat by email or name
            #email_name = chat_name
            #chat_id = ChatUtils.get_existing_one_on_one_chat_id(email_name)
            #if chat_id is not None:
            #    return ChatUtils.send_chat_with_id(chat_id, messageHTML)

            # Not finding in existing chat, search for matching users
            print(f"Searching for users with name '{chat_name}'...", flush=True)
            # Search for users by name
            users = UserUtils.get_users_by_name(chat_name) 
            #print(f"chat.py: users = UserUtils.get_users_by_name(chat_name) = {users}", flush=True) 
            if not users:
                return {
                    "status": "error",
                    "message": f"No users found matching the name '{chat_name}'."
                }
            elif len(users) == 1:
                # If there's only one match, use it automatically
                print(f"Found one user. Sending message directly...", flush=True)
                print(f"Found one user: {users[0]['id']} {users[0]['displayName']} ({users[0]['userPrincipalName']})", flush=True)
                email = users[0].get("userPrincipalName") or users[0].get("mail")
                user_id = users[0].get("id")
                chat_id = ChatUtils.create_one_on_one_chat_id(user_id)
                if chat_id is None:
                    return {
                        "status": "error",
                        "message": f"Could not create chat with user '{email}'."
                    }
                else:
                    return ChatUtils.send_chat_with_id(chat_id, messageHTML,image_path)
            else:
                if user_index is not None and 0 <= user_index < len(users):
                    selected_user = users[user_index-1]  # Adjust for 1-based index
                    print(f"Selected user: {selected_user['displayName']} ({selected_user['userPrincipalName']})", flush=True)
                    email = selected_user.get("userPrincipalName") or selected_user.get("mail")
                    user_id = selected_user.get("id")
                    chat_id = ChatUtils.create_one_on_one_chat_id(user_id)
                    if chat_id is None:
                        return {
                            "status": "error",
                            "message": f"Could not create chat with user '{email}'."
                        }
                    else:
                        return ChatUtils.send_chat_with_id(chat_id, messageHTML,image_path)
                else:
                    for i, user in enumerate(users):
                        print(f"{i+1}. {user['displayName']} ({user['userPrincipalName']}) {user['jobTitle']}", flush=True)
                    return {
                        "status": "multiple_matches",
                        "message": f"Multiple users found matching the name '{chat_name}'. Please select one.",
                        "users": users
                    }
        else:
            return {
                "status": "error",
                "message": "Invalid type. Please use 'myself', 'group', or 'person'."
            }

    @staticmethod
    def create_one_on_one_chat_id(user_id: str):
        access_token = AuthUtils.login()
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json"
        }

        print(f"Creating 1:1 chat with user '{user_id}'...", flush=True)
        current_user_id = UserUtils.get_current_user().get("id")
        if not current_user_id:
            print("Failed to get current user ID", flush=True)
            return None

        # if the user ID is the same as the current user, then return "self"
        if current_user_id == user_id:
            return "self"

        payload = {
            "chatType": "oneOnOne",
            "members": [
                {
                    "@odata.type": "#microsoft.graph.aadUserConversationMember",
                    "roles": ["owner"],
                    "user@odata.bind": f"https://graph.microsoft.com/v1.0/users/{current_user_id}"
                },
                {
                    "@odata.type": "#microsoft.graph.aadUserConversationMember",
                    "roles": ["owner"],
                    "user@odata.bind": f"https://graph.microsoft.com/v1.0/users/{user_id}"
                }
            ]
        }
        
        create_chat_url = "https://graph.microsoft.com/v1.0/chats"
        create_resp = requests.post(create_chat_url, headers=headers, json=payload)
        if create_resp.status_code in (201, 200):
            print(f"1:1 chat created successfully with user '{user_id}'.", flush=True)
            return create_resp.json().get("id")
        else:
            print(f"Error creating 1:1 chat: {create_resp.text}", flush=True)
            return None

    @staticmethod
    def get_existing_one_on_one_chat_id(email_name: str):
        access_token = AuthUtils().login()
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json"
        }

        # List top 1:1 chats and look for a chat with the user
        print(f"Searching for 1:1 chat with user '{email_name}'...", flush=True)
        chats_url = "https://graph.microsoft.com/v1.0/me/chats?$filter=chatType eq 'oneOnOne'"
        search_count = 0  # Initialize counter, adding to limit processing to 20 members for performance
        # Allow max_search_count to be set as a class variable or default to 20
        max_search_count = 20

        while chats_url and search_count <= max_search_count:
            response = requests.get(chats_url, headers=headers)
            if response.status_code != 200:
                print(f"Error retrieving 1:1 chats (HTTP {response.status_code}): {response.text}", flush=True)
                return None
            for chat in response.json().get("value", []):
                members = ChatUtils.get_chat_members(chat["id"])
                if members:
                    for member in members:
                        search_count += 1
                        if search_count > max_search_count:
                            print(f"More than {max_search_count} members processed, stopping further processing.", flush=True)
                            break
                        # Check both 'email' and 'userPrincipalName'                        
                        # Get member values safely, ensuring they're never None
                        member_email = member.get("email", "") or ""
                        member_upn = member.get("userPrincipalName", "") or ""
                        member_display = member.get("displayName", "") or ""
                        print(f"Checking member: {member_display} ({member_email}, {member_upn})", flush=True)
                        if (
                            member_email.lower() == email_name.lower() or
                            member_upn.lower() == email_name.lower() or
                            member_display.lower() == email_name.lower()
                        ):
                            return chat["id"]                            
            chats_url = response.json().get("@odata.nextLink")

        print(f"No 1:1 chat found with user '{email_name}' after checking {search_count} members.", flush=True)
        return None